/*
    This file is a part of Musty, a simple Molecular Dynamics simulation
    package.
    
    Copyright (C) 2005 Jori Liesenborgs

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mdsimulation.h"
#include "interparticleforce.h"
#include "constants.h"
#include "uniformdistribution.h"
#include <stdio.h>

MDSimulation::MDSimulation(int numparticles)
{
	numpart = numparticles;
	particles = new Particle[numpart];
	forces = new Vector3D[numpart];
	prevforces = new Vector3D[numpart];
	particlestep = new Vector3D[numpart];
}

MDSimulation::~MDSimulation()
{
	delete [] particles;
	delete [] forces;
	delete [] prevforces;
	delete [] particlestep;
}

bool MDSimulation::Run(InterParticleForce &force,real_t temp,real_t boxsize,real_t timestep,
                       real_t lengthscale,real_t massscale,real_t energyscale,int equilibsteps)
{
	// set units
	
	M = massscale;
	L = lengthscale;
	E = energyscale;
	T = SQRT(M/E)*L;
	F = E/L;
	kB = 1.381e-23/E;

	// store some parameters
	temperature = temp;
	boxlength = boxsize;
	tstep = timestep;
	
	InitializeParticles(particles,numpart);
	for (int i = 0 ; i < 0 ; i++)
		particlestep[i] = Vector3D(0,0,0);
	
	ZeroMomentum();
	RescaleVelocities();	
	WrapParticles();

	equilib = false;
	bool done = false;
	
	if (!OnStart())
		return false;
	
	if (!OnStep(&done))
	{
		OnStop(true);
		return false;
	}

	if (done)
	{
		OnStop(false);
		return true;
	}
	
	FirstStep(force);
	WrapParticles();
	
	RescaleVelocities();
	int count = 0;
	while (!equilib && !done)
	{
		for (int i = 0 ; !done && i < 20 ; i++)
		{
			if (!OnStep(&done))
			{
				OnStop(true);
				return false;
			}
			if (!done)
			{
				Step(force);
				WrapParticles();
			}
			count++;
			if ((count % 100) == 0)
				printf("count = %d\n",count);
		}
		
		if (!done)
		{
			real_t lambda = RescaleVelocities();

			printf("lambda = %g\n",(double)lambda);
		}
		if (count >= equilibsteps)
			equilib = true;
	}

	while (!done)
	{
		if (!OnStep(&done))
		{
			OnStop(true);
			return false;
		}
		if (!done)
		{
			Step(force);
			WrapParticles();
		}
	}
	
	OnStop(false);
	
	return true;
}

void MDSimulation::InitializeParticles(Particle *p,int numparticles)
{
	InitializeMasses(p,numparticles);
	InitializeRandomVelocities(p,numparticles);
	InitializeFCCPositions(p,numparticles);
}

void MDSimulation::InitializeFCCPositions(Particle *p,int numparticles)
{
	int i;
	int numblocks = (int)(POW((((real_t)numparticles)/4.0),1.0/3.0)+0.5);
	int xpos,ypos,zpos;
	real_t len = GetBoxSize()/((real_t)(numblocks));

	xpos = 0;
	ypos = 0;
	zpos = 0;
	
	for (i = 0 ; i < numparticles ; i++ )
	{
		real_t x,y,z;

		x = (((real_t)xpos))*len;
		y = (((real_t)ypos))*len;
		z = (((real_t)zpos))*len;
		
		p[i].SetPosition(Vector3D(x,y,z));
			
		i++;
		if (i < numparticles)
			p[i].SetPosition(Vector3D(x+len/2.0,y+len/2.0,z));
		i++;
		if (i < numparticles)
			p[i].SetPosition(Vector3D(x,y+len/2.0,z+len/2.0));
		i++;
		if (i < numparticles)
			p[i].SetPosition(Vector3D(x+len/2.0,y,z+len/2.0));
		
		xpos++;
		if (xpos == numblocks)
		{
			xpos = 0;
			ypos++;
			if (ypos == numblocks)
			{
				ypos = 0;
				zpos++;
			}
		}
	}}

void MDSimulation::InitializeRandomVelocities(Particle *p,int numparticles)
{
	int i;
	real_t avgenergy = (((real_t)(numparticles-1))/((real_t)numparticles))*1.5*Get_kB()*GetTemperature();
	UniformDistribution unit(-1,1);

	for (i = 0 ; i < numparticles ; i++)
	{
		real_t x,y,z;

		x = unit.PickNumber();
		y = unit.PickNumber();
		z = unit.PickNumber();
		
		Vector3D unitvect = Vector3D(x,y,z);

		unitvect /= unitvect.GetLength();

		real_t v = SQRT(avgenergy*2.0/p[i].GetMass());

		p[i].SetVelocity(unitvect*v);
	}
}

void MDSimulation::InitializeMasses(Particle *p,int numparticles)
{
	int i;
	for (i = 0 ; i < numparticles ; i++)
		p[i].SetMass(1.0);
}

void MDSimulation::ZeroMomentum()
{
	Vector3D totalmomentum(0,0,0);
	int i;

	for (i = 0 ; i < numpart ; i++)
		totalmomentum += particles[i].GetVelocity()*particles[i].GetMass();

	Vector3D avgmomentum = totalmomentum/((real_t)numpart);

	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D vel = particles[i].GetVelocity();

		vel -= avgmomentum/particles[i].GetMass();
		particles[i].SetVelocity(vel);
	}
}

real_t MDSimulation::RescaleVelocities()
{
	real_t mv2 = 0;
	int i;

	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D vel = particles[i].GetVelocity();
		mv2 += particles[i].GetMass()*vel.GetLength2();
	}

	if (mv2 == 0)
		return -1.0;

	real_t lambda = SQRT((((real_t)numpart-1)*3.0*kB*temperature)/mv2);
	
	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D vel = particles[i].GetVelocity();
		
		vel *= lambda;
		particles[i].SetVelocity(vel);
	}

	return lambda;
}

void MDSimulation::WrapParticles()
{
	int i;

	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D pos;
		real_t x,y,z;

		pos = particles[i].GetPosition();
		x = pos.GetX();
		y = pos.GetY();
		z = pos.GetZ();

		real_t a = x/boxlength;
		real_t b = y/boxlength;
		real_t c = z/boxlength;

		if (a > 1.0)
			x -= ((real_t)((int64_t)(a+0.5)))*boxlength;
		else if (a < 0)
			x += ((real_t)((int64_t)(-a+1.5)))*boxlength;
		if (b > 1.0)
			y -= ((real_t)((int64_t)(b+0.5)))*boxlength;
		else if (b < 0)
			y += ((real_t)((int64_t)(-b+1.5)))*boxlength;
		if (c > 1.0)
			z -= ((real_t)((int64_t)(c+0.5)))*boxlength;
		else if (c < 0)
			z += ((real_t)((int64_t)(-c+1.5)))*boxlength;
		
		while (x < 0) { x += boxlength; }
		while (x > boxlength) { x -= boxlength; }
		while (y < 0) { y += boxlength; }
		while (y > boxlength) { y -= boxlength; }
		while (z < 0) { z += boxlength; }
		while (z > boxlength) { z -= boxlength; }

		particles[i].SetPosition(Vector3D(x,y,z));
	}
}

void MDSimulation::FirstStep(InterParticleForce &force)
{
	int i;
	
	// update positions
	
	CalculateForces(force);
	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D oldpos = particles[i].GetPosition();
		Vector3D newpos = particles[i].GetPosition()+tstep*particles[i].GetVelocity()+(tstep*tstep/2.0)*forces[i]/particles[i].GetMass();
		particles[i].SetPosition(newpos);
		particlestep[i] = newpos-oldpos;
	}
	
	// update velocities
	
	Vector3D *tmp = prevforces;
	
	prevforces = forces;
	forces = tmp;
	
	CalculateForces(force);
	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D newvel = particles[i].GetVelocity()+tstep*0.5*(forces[i]+prevforces[i])/particles[i].GetMass();
		particles[i].SetVelocity(newvel);
	}
}

void MDSimulation::Step(InterParticleForce &force)
{
	int i;

	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D oldpos = particles[i].GetPosition();
		Vector3D newpos = particles[i].GetPosition()+tstep*particles[i].GetVelocity()+(tstep*tstep/2.0)*forces[i]/particles[i].GetMass();
		particles[i].SetPosition(newpos);
		particlestep[i] = newpos-oldpos;
	}
	
	// update velocities
	
	Vector3D *tmp = prevforces;
	
	prevforces = forces;
	forces = tmp;
	
	CalculateForces(force);
	for (i = 0 ; i < numpart ; i++)
	{
		Vector3D newvel = particles[i].GetVelocity()+tstep*0.5*(forces[i]+prevforces[i])/particles[i].GetMass();
		particles[i].SetVelocity(newvel);
	}
}

void MDSimulation::CalculateForces(InterParticleForce &force)
{
	int i,j;
	
	for (i = 0 ; i < numpart ; i++)
		forces[i] = Vector3D(0,0,0);
	
	for (i = 0 ; i < numpart ; i++)
	{
		for (j = i+1 ; j < numpart ; j++)
		{
			Vector3D mindiff = GetInterParticleVector(i,j);
			real_t mindist = mindiff.GetLength();

			Vector3D unitvect = mindiff/mindist;
			real_t forcesize = force.GetForce(mindist,particles[i].GetMass(),particles[j].GetMass());
			Vector3D vectforce = unitvect*forcesize;

			forces[i] += vectforce;
			forces[j] -= vectforce;
		}
	}
}

bool MDSimulation::OnStart()
{
	return true;
}

bool MDSimulation::OnStep(bool *stop)
{
	return true;
}

void MDSimulation::OnStop(bool error)
{
}

Vector3D MDSimulation::GetInterParticleVector(int i,int j) const
{
	real_t Lx,Ly,Lz;
	Vector3D diff = particles[j].GetPosition()-particles[i].GetPosition();
	Vector3D origpos = particles[j].GetPosition();
	Vector3D part1pos = particles[i].GetPosition();
	real_t mindist = diff.GetLength2();
	Vector3D mindiff = diff;
			
	for (Lx = -1 ; Lx <= 1 ; Lx++)
	{
		for (Ly = -1 ; Ly <= 1 ; Ly++)
		{
			for (Lz = -1 ; Lz <= 1 ; Lz++)
			{
				Vector3D newpos = origpos+Vector3D(Lx,Ly,Lz)*boxlength;
				diff = newpos-part1pos;

				real_t difflen2 = diff.GetLength2();
				if (difflen2 < mindist)
				{
					mindist = difflen2;
					mindiff = diff;
				}
			}
		}
	}
	return mindiff;
}


