/*
    This file is a part of Musty, a simple Molecular Dynamics simulation
    package.
    
    Copyright (C) 2005 Jori Liesenborgs

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MDSIMULATION_H

#define MDSIMULATION_H

#include "particle.h"

class InterParticleForce;

class MDSimulation
{
public:
	MDSimulation(int numparticles);
	virtual ~MDSimulation();

	// timestep and boxsize should be in the specified units
	bool Run(InterParticleForce &force,real_t T,real_t boxsize,real_t timestep,
	         real_t lengthscale = 1.0,real_t massscale = 1.0,real_t energyscale = 1.0,int equilibsteps = 2500);
protected:
	real_t GetLengthUnit() const							{ return L; }
	real_t GetTimeUnit() const							{ return T; }
	real_t GetTimeStep() const							{ return tstep; }
	real_t GetTemperature() const							{ return temperature; }
	real_t Get_kB() const								{ return kB; }
	real_t GetBoxSize() const							{ return boxlength; }
	const Particle *GetParticles() const						{ return particles; }
	const Vector3D *GetParticleSteps() const					{ return particlestep; }
	int GetNumberOfParticles() const						{ return numpart; }
	bool InEquilibrium() const							{ return equilib; }
	Vector3D GetInterParticleVector(int i,int j) const;
	virtual void InitializeParticles(Particle *particles,int numparticles);
	void InitializeFCCPositions(Particle *particles,int numparticles);
	void InitializeRandomVelocities(Particle *particles,int numparticles);
	void InitializeMasses(Particle *particles,int numparticles);
	virtual bool OnStart();
	virtual bool OnStep(bool *stop);
	virtual void OnStop(bool error);
private:
	void ZeroMomentum();
	real_t RescaleVelocities();
	void WrapParticles();
	void CalculateForces(InterParticleForce &force);
	void FirstStep(InterParticleForce &force);
	void Step(InterParticleForce &force);
	
	Particle *particles;
	Vector3D *particlestep;
	Vector3D *forces,*prevforces;
	int numpart;
	real_t temperature;
	real_t tstep,boxlength;
	bool equilib;

	// units
	real_t M,L,T,E,F,kB;
};

#endif // MDSIMULATION_H

